#ifndef MY_ABC_HERE
#define MY_ABC_HERE
#endif
/* chmod -- change permission modes of files
   Copyright (C) 1989-2018 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* Written by David MacKenzie <djm@gnu.ai.mit.edu> */

#include <config.h>
#include <stdio.h>
#include <getopt.h>
#include <sys/types.h>

#ifdef MY_ABC_HERE
#   include <string.h>
#   include <ctype.h>
#   include <synoacl/synoacl.h>
#   define DEFAULT_ACL_INDEX_SIZE 10
#endif

#include "system.h"
#include "dev-ino.h"
#include "die.h"
#include "error.h"
#include "filemode.h"
#include "ignore-value.h"
#include "modechange.h"
#include "quote.h"
#include "root-dev-ino.h"
#include "xfts.h"

/* The official name of this program (e.g., no 'g' prefix).  */
#define PROGRAM_NAME "chmod"

#define AUTHORS \
  proper_name ("David MacKenzie"), \
  proper_name ("Jim Meyering")

enum Change_status
{
  CH_NOT_APPLIED,
  CH_SUCCEEDED,
  CH_FAILED,
  CH_NO_CHANGE_REQUESTED
};

enum Verbosity
{
  /* Print a message for each file that is processed.  */
  V_high,

  /* Print a message for each file whose attributes we change.  */
  V_changes_only,

  /* Do not be verbose.  This is the default. */
  V_off
};

/* The desired change to the mode.  */
static struct mode_change *change;

#ifdef MY_ABC_HERE
/* The list for ACL index operation */
static int *acl_index;

/* If true, keep file be acl mode even if POSIX mode is given */
static bool force_keep_acl;
#endif //MY_ABC_HERE

/* The initial umask value, if it might be needed.  */
static mode_t umask_value;

/* If true, change the modes of directories recursively. */
static bool recurse;

/* If true, force silence (suppress most of error messages). */
static bool force_silent;

/* If true, diagnose surprises from naive misuses like "chmod -r file".
   POSIX allows diagnostics here, as portable code is supposed to use
   "chmod -- -r file".  */
static bool diagnose_surprises;

/* Level of verbosity.  */
static enum Verbosity verbosity = V_off;

/* Pointer to the device and inode numbers of '/', when --recursive.
   Otherwise NULL.  */
static struct dev_ino *root_dev_ino;

/* For long options that have no equivalent short option, use a
   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
enum
{
  NO_PRESERVE_ROOT = CHAR_MAX + 1,
  PRESERVE_ROOT,
  REFERENCE_FILE_OPTION
};

static struct option const long_options[] =
{
  {"changes", no_argument, NULL, 'c'},
  {"recursive", no_argument, NULL, 'R'},
  {"no-preserve-root", no_argument, NULL, NO_PRESERVE_ROOT},
  {"preserve-root", no_argument, NULL, PRESERVE_ROOT},
  {"quiet", no_argument, NULL, 'f'},
  {"reference", required_argument, NULL, REFERENCE_FILE_OPTION},
  {"silent", no_argument, NULL, 'f'},
  {"verbose", no_argument, NULL, 'v'},
  {GETOPT_HELP_OPTION_DECL},
  {GETOPT_VERSION_OPTION_DECL},
  {NULL, 0, NULL, 0}
};

/* Return true if the chmodable permission bits of FILE changed.
   The old mode was OLD_MODE, but it was changed to NEW_MODE.  */

#ifdef MY_ABC_HERE

static void
parse_synoacl_index(char* index_list, char** p_mode)
{
  int n_acl_index = DEFAULT_ACL_INDEX_SIZE;
  int index_num = 0;
  int index = 0;
  int *mem_swap = NULL;
  char *pch = NULL;

  (*p_mode)[2] = '\0';

  acl_index = malloc(sizeof(int) * n_acl_index);
  if ( NULL == acl_index ) {
    error(0, ENOMEM, _("out of memory"));
    usage(EXIT_FAILURE);
  }
  memset(acl_index, 0xFF, sizeof(int) * n_acl_index);

  for(pch = index_list; '\0' != *pch; pch++) {
    if(*pch == ',') {
      *pch = ' ';
    }
  };

  pch = index_list;
  while (ISDIGIT(*pch) || ' ' == *pch) {
    index = strtol(pch, &pch, 10);

    if ( 0 > index || SYNOSDK_ACL_MAX_ACE_COUNT <= index ) {
      error(0, EPERM, _("index is out of range"));
      usage(EXIT_FAILURE);
    }

    index_num += 1;
    if ( (index_num+1) > n_acl_index ) {
      n_acl_index *= 2;
      mem_swap = realloc(acl_index, n_acl_index);
      if ( NULL == mem_swap ) {
        error(0, ENOMEM, _("out of memory"));
        usage(EXIT_FAILURE);
      }
      acl_index = mem_swap;
      mem_swap = NULL;
    }
    acl_index[index_num-1] = index;
  }

  if ( index_num == 0 ) {
    error(0, EPERM, _("index number should be given"));
    usage(EXIT_FAILURE);
  }
  acl_index[index_num] = -1;
}

static void
parse_synoacl_acl(char* acl_list, char** p_mode)
{
  int mem_size = 0;
  char *mem_swap = NULL;

  mem_size = strlen(*p_mode) +1 + strlen(acl_list) +1;
  mem_swap = realloc(*p_mode, mem_size);
  if ( NULL == mem_swap ) {
    error(0, ENOMEM, _("out of memory"));
    usage(EXIT_FAILURE);
  }
  snprintf(mem_swap+strlen(mem_swap), mem_size-strlen(mem_swap), "%s", acl_list);
  *p_mode = mem_swap;
}

/*
 * this function does normalize mode string
 *
 * 1. [-+]a ACL,ACL... => [-+a]ACL,ACL...
 * 2. -a# idx,idx... => -a
 * 3. =a# idx ACL,ACL... => =aACL,ACL...
 *
 * acl_index is filled if there is '#' specifier.
 */
static void
parse_synoacl_arguments(int argc, char **argv,
    char** p_mode, bool* p_set_acl, bool* p_is_acl_opindex)
{
  size_t mode_len = 0;
  char* mode = NULL;
  char* acl_mode  = NULL;
  char operator = '\0';
  bool need_parse_index = false;
  bool need_parse_acl = false;

  if (!argv || !p_mode || !p_set_acl || !p_is_acl_opindex) {
      error(0, ENOMEM, _("out of memory"));
      usage(EXIT_FAILURE);
  }
  if (NULL == (mode = *p_mode)) {
    return;
  }
  if (2 > (mode_len = strlen(mode))) {
    return;
  }

  if (strncmp(mode, "-a", 2) && strncmp(mode, "+a", 2) && strncmp(mode, "=a", 2)) {
    return;
  }

  operator = mode[0];
  need_parse_index = *p_is_acl_opindex = ('#' == mode[2]); // mode[2] == '\0' or not

  // +a  cannot be used with #
  switch(operator) {
    case '-':
      need_parse_acl = !need_parse_index;
      break;
    case '=':
      need_parse_acl = true;
      if (!need_parse_index) {
        error(0, 0, _("invalid mode: %s"), quote (mode));
        usage(EXIT_FAILURE);
      }
      break;
    case '+':
      need_parse_acl = true;
      if (need_parse_index) {
        error(0, 0, _("invalid mode: %s"), quote (mode));
        usage(EXIT_FAILURE);
      }
      break;
    default:
      error(0, 0, _("invalid mode: %s"), quote (mode));
      usage(EXIT_FAILURE);
  }

  acl_mode = strdup(mode);
  if (NULL == acl_mode) {
    error(0, ENOMEM, _("out of memory"));
    usage(EXIT_FAILURE);
  }

  if (need_parse_index) {
    if ( optind >=argc ) {
      error(0, 0, _("missing operand after %s"), mode);
      usage(EXIT_FAILURE);
    }
    parse_synoacl_index(argv[optind], &acl_mode);
    optind++;
  }

  if (need_parse_acl) {
    if ( optind >=argc ) {
      error(0, 0, _("missing operand after %s"), mode);
      usage(EXIT_FAILURE);
    }
    parse_synoacl_acl(argv[optind], &acl_mode);
    optind++;
  }

  *p_mode = acl_mode;
  *p_set_acl = true;
}
#endif //MY_ABC_HERE

static bool
mode_changed (int dir_fd, char const *file, char const *file_full_name,
              mode_t old_mode, mode_t new_mode)
{
  if (new_mode & (S_ISUID | S_ISGID | S_ISVTX))
    {
      /* The new mode contains unusual bits that the call to chmod may
         have silently cleared.  Check whether they actually changed.  */

      struct stat new_stats;

      if (fstatat (dir_fd, file, &new_stats, 0) != 0)
        {
          if (! force_silent)
            error (0, errno, _("getting new attributes of %s"),
                   quoteaf (file_full_name));
          return false;
        }

      new_mode = new_stats.st_mode;
    }

  return ((old_mode ^ new_mode) & CHMOD_MODE_BITS) != 0;
}

/* Tell the user how/if the MODE of FILE has been changed.
   CHANGED describes what (if anything) has happened. */

static void
describe_change (const char *file, mode_t old_mode, mode_t mode,
                 enum Change_status changed)
{
  char perms[12];       /* "-rwxrwxrwx" ls-style modes. */
  char old_perms[12];
  const char *fmt;

  if (changed == CH_NOT_APPLIED)
    {
      printf (_("neither symbolic link %s nor referent has been changed\n"),
              quoteaf (file));
      return;
    }

  strmode (mode, perms);
  perms[10] = '\0';     /* Remove trailing space.  */

  strmode (old_mode, old_perms);
  old_perms[10] = '\0';     /* Remove trailing space.  */

  switch (changed)
    {
    case CH_SUCCEEDED:
      fmt = _("mode of %s changed from %04lo (%s) to %04lo (%s)\n");
      break;
    case CH_FAILED:
      fmt = _("failed to change mode of %s from %04lo (%s) to %04lo (%s)\n");
      break;
    case CH_NO_CHANGE_REQUESTED:
      fmt = _("mode of %s retained as %04lo (%s)\n");
      printf (fmt, quoteaf (file),
              (unsigned long int) (mode & CHMOD_MODE_BITS), &perms[1]);
      return;
    default:
      abort ();
    }
  printf (fmt, quoteaf (file),
          (unsigned long int) (old_mode & CHMOD_MODE_BITS), &old_perms[1],
          (unsigned long int) (mode & CHMOD_MODE_BITS), &perms[1]);
}

/* Change the mode of FILE.
   Return true if successful.  This function is called
   once for every file system object that fts encounters.  */

static bool
process_file (FTS *fts, FTSENT *ent)
{
  char const *file_full_name = ent->fts_path;
  char const *file = ent->fts_accpath;
  const struct stat *file_stats = ent->fts_statp;
#ifdef MY_ABC_HERE
  SYNO_ACL *old_acl = NULL;
  SYNO_ACL *new_acl = get_acl_from_mode_change(change);
#endif // MY_ABC_HERE
  mode_t old_mode IF_LINT ( = 0);
  mode_t new_mode IF_LINT ( = 0);
  bool ok = true;
  bool chmod_succeeded = false;
#ifdef MY_ABC_HERE
  bool is_acl_file = false;

  if (acl_index && new_acl && 1 < new_acl->counts) {
    error(0, EPERM, _("acl index operation cannot be performed \
                with multiple aces"));
    usage(EXIT_FAILURE);
  }
#endif //MY_ABC_HERE

  switch (ent->fts_info)
    {
    case FTS_DP:
      return true;

    case FTS_NS:
      /* For a top-level file or directory, this FTS_NS (stat failed)
         indicator is determined at the time of the initial fts_open call.
         With programs like chmod, chown, and chgrp, that modify
         permissions, it is possible that the file in question is
         accessible when control reaches this point.  So, if this is
         the first time we've seen the FTS_NS for this file, tell
         fts_read to stat it "again".  */
      if (ent->fts_level == 0 && ent->fts_number == 0)
        {
          ent->fts_number = 1;
          fts_set (fts, ent, FTS_AGAIN);
          return true;
        }
      if (! force_silent)
        error (0, ent->fts_errno, _("cannot access %s"),
               quoteaf (file_full_name));
      ok = false;
      break;

    case FTS_ERR:
      if (! force_silent)
        error (0, ent->fts_errno, "%s", quotef (file_full_name));
      ok = false;
      break;

    case FTS_DNR:
      if (! force_silent)
        error (0, ent->fts_errno, _("cannot read directory %s"),
               quoteaf (file_full_name));
      ok = false;
      break;

    case FTS_SLNONE:
      if (! force_silent)
        error (0, 0, _("cannot operate on dangling symlink %s"),
               quoteaf (file_full_name));
      ok = false;
      break;

    case FTS_DC:        /* directory that causes cycles */
      if (cycle_warning_required (fts, ent))
        {
          emit_cycle_warning (file_full_name);
          return false;
        }
      break;

    default:
      break;
    }

  if (ok && ROOT_DEV_INO_CHECK (root_dev_ino, file_stats))
    {
      ROOT_DEV_INO_WARN (file_full_name);
      /* Tell fts not to traverse into this hierarchy.  */
      fts_set (fts, ent, FTS_SKIP);
      /* Ensure that we do not process "/" on the second visit.  */
      ignore_value (fts_read (fts));
      return false;
    }

  if (ok)
    {
      old_mode = file_stats->st_mode;
#ifdef MY_ABC_HERE
      is_acl_file = IS_FILE_SUPPORT_ACL(file_full_name);

      if (!is_acl_file && !force_keep_acl && (new_acl || acl_index) ) {
        if (! force_silent)
          error (-1, EPERM, _("%s is linux mode"), quote(file_full_name));
        return false;
      }

      if (is_acl_file && (new_acl || acl_index)) {
        SYNOACLGet(file_full_name, -1, SYNOACL_SELF, &old_acl);

        new_acl = acl_adjust(old_acl, new_acl, acl_index, change);
        if ( !new_acl ) {
            error(0, errno, _("failed to merge acl"));
            return false;
        }

      } else if (is_acl_file && force_keep_acl) {
        SYNOACLGet(file_full_name, -1, SYNOACL_SELF, &old_acl);

        new_acl = trf_mode_to_acl(change, file_stats);
        if( NULL == new_acl ){
            error(-1, errno, _("out of memory"));
            return false;
        }

        new_acl = acl_adjust(old_acl, new_acl, acl_index, change);
        if ( NULL == new_acl ) {
            error(0, errno, _("failed to merge acl(-k)"));
            return false;
        }

      } else
#endif // MY_ABC_HERE
      new_mode = mode_adjust (old_mode, S_ISDIR (old_mode) != 0, umask_value,
                              change, NULL);

      if (! S_ISLNK (old_mode))
        {
#ifdef MY_ABC_HERE
          if (is_acl_file && new_acl) {
            if (0 == SYNOACLSet(file_full_name, -1, new_acl)) {
                chmod_succeeded = true;
            }
            SYNOACLFree(new_acl);
            new_acl = NULL;
          } else
#endif // MY_ABC_HERE
          if (chmodat (fts->fts_cwd_fd, file, new_mode) == 0)
            chmod_succeeded = true;
          else
            {
              if (! force_silent)
                error (0, errno, _("changing permissions of %s"),
                       quoteaf (file_full_name));
              ok = false;
            }
        }
    }

  if (verbosity != V_off)
    {
#ifdef MY_ABC_HERE
      bool changed = false;
      if ( is_acl_file ) {
        changed = chmod_succeeded;
      } else {
        changed = (chmod_succeeded
                    && mode_changed (fts->fts_cwd_fd, file, file_full_name,
                                       old_mode, new_mode));
      }
#else
      bool changed = (chmod_succeeded
                      && mode_changed (fts->fts_cwd_fd, file, file_full_name,
                                       old_mode, new_mode));
#endif // MY_ABC_HERE

      if (changed || verbosity == V_high)
        {
          enum Change_status ch_status =
            (!ok ? CH_FAILED
             : !chmod_succeeded ? CH_NOT_APPLIED
             : !changed ? CH_NO_CHANGE_REQUESTED
             : CH_SUCCEEDED);
          describe_change (file_full_name, old_mode, new_mode, ch_status);
        }
    }

  if (chmod_succeeded && diagnose_surprises)
    {
      mode_t naively_expected_mode =
        mode_adjust (old_mode, S_ISDIR (old_mode) != 0, 0, change, NULL);
      if (new_mode & ~naively_expected_mode)
        {
          char new_perms[12];
          char naively_expected_perms[12];
          strmode (new_mode, new_perms);
          strmode (naively_expected_mode, naively_expected_perms);
          new_perms[10] = naively_expected_perms[10] = '\0';
          error (0, 0,
                 _("%s: new permissions are %s, not %s"),
                 quotef (file_full_name),
                 new_perms + 1, naively_expected_perms + 1);
          ok = false;
        }
    }

  if ( ! recurse)
    fts_set (fts, ent, FTS_SKIP);

  return ok;
}

/* Recursively change the modes of the specified FILES (the last entry
   of which is NULL).  BIT_FLAGS controls how fts works.
   Return true if successful.  */

static bool
process_files (char **files, int bit_flags)
{
  bool ok = true;

  FTS *fts = xfts_open (files, bit_flags, NULL);

  while (1)
    {
      FTSENT *ent;

      ent = fts_read (fts);
      if (ent == NULL)
        {
          if (errno != 0)
            {
              /* FIXME: try to give a better message  */
              if (! force_silent)
                error (0, errno, _("fts_read failed"));
              ok = false;
            }
          break;
        }

      ok &= process_file (fts, ent);
    }

  if (fts_close (fts) != 0)
    {
      error (0, errno, _("fts_close failed"));
      ok = false;
    }
#ifdef MY_ABC_HERE
  SYNOACLFree(get_acl_from_mode_change(change));
  if( acl_index ) {
    free(acl_index);
    acl_index = NULL;
  }
#endif // MY_ABC_HERE

  return ok;
}

void
usage (int status)
{
  if (status != EXIT_SUCCESS)
    emit_try_help ();
  else
    {
      printf (_("\
Usage: %s [OPTION]... MODE[,MODE]... FILE...\n\
  or:  %s [OPTION]... OCTAL-MODE FILE...\n\
  or:  %s [OPTION]... --reference=RFILE FILE...\n\
"),
              program_name, program_name, program_name);
      fputs (_("\
Change the mode of each FILE to MODE.\n\
With --reference, change the mode of each FILE to that of RFILE.\n\
\n\
"), stdout);
      fputs (_("\
  -c, --changes          like verbose but report only when a change is made\n\
  -f, --silent, --quiet  suppress most error messages\n\
  -v, --verbose          output a diagnostic for every file processed\n\
"), stdout);
      fputs (_("\
      --no-preserve-root  do not treat '/' specially (the default)\n\
      --preserve-root    fail to operate recursively on '/'\n\
"), stdout);
      fputs (_("\
      --reference=RFILE  use RFILE's mode instead of MODE values\n\
"), stdout);
      fputs (_("\
  -R, --recursive        change files and directories recursively\n\
"), stdout);
      fputs (HELP_OPTION_DESCRIPTION, stdout);
      fputs (VERSION_OPTION_DESCRIPTION, stdout);
      fputs (_("\
\n\
Each MODE is of the form '[ugoa]*([-+=]([rwxXst]*|[ugo]))+|[-+=][0-7]+'.\n\
"), stdout);
#ifdef MY_ABC_HERE
      fputs (_("\
\n\
  -k                     force keep the file in acl mode even if linux perm \n\
                         is given. the mode NNN will be tranformed to ACLs \n\
                         700 = u:<user of file>:allow:MRWX:fd \n\
                         060 = g:<group of file>:allow:MRW:fd\n\
                         003 = a::allow:MWX:fd\n\
SYNOACL Operations:\n\
  [-+]a <ACL[,ACL...]>       add/remove specific ACEs\n\
  -a# <index[,index...]>     remove the ACEs with spcific indices\n\
  =a# <index[,index...]> ACL replace specific ACEs with the given ACL\n\
\n\
ACL Format:\n\
  <Object Type>:<Object name>:<Access type>:<Permissions>:<Inheritance>\n\
\n\
  Object:      "SYNOSDK_ACL_STR_USER"(u),"SYNOSDK_ACL_STR_GROUP"(g),"
               SYNOSDK_ACL_STR_EVERYONE"(a),"SYNOSDK_ACL_STR_OWNER"(o),\n\
               "SYNOSDK_ACL_STR_AUTHENTICATEDUSER"(l),"
               SYNOSDK_ACL_STR_SYSTEM"(s)\n\
               Object name can be empty if object type is not user/group\n\
  Access:      allow/deny\n\
  Permission:  rwxpdDaARWcCo, MRWX, F\n\
               MR = raRc, MW = wpDAW, MX = x, F=rwxpdDaARWcCo\n\
  Inheritance: fdin; Can be empty, default: fd\n\
\n\
*Notice: SYNOACL operations will do normalization first, which means\n\
         duplicated ACEs in ACL will be removed and ACEs with same object, \n\
         same access type and same inheritance will be merged.\n\
"), stdout);
#endif // MY_ABC_HERE
      emit_ancillary_info (PROGRAM_NAME);
    }
  exit (status);
}

/* Parse the ASCII mode given on the command line into a linked list
   of 'struct mode_change' and apply that to each file argument. */

int
main (int argc, char **argv)
{
  char *mode = NULL;
  size_t mode_len = 0;
  size_t mode_alloc = 0;
  bool ok;
  bool preserve_root = false;
#ifdef MY_ABC_HERE
  bool set_acl = false;
  bool is_acl_opindex = false;
#endif // MY_ABC_HERE
  char const *reference_file = NULL;
  int c;

  initialize_main (&argc, &argv);
  set_program_name (argv[0]);
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  atexit (close_stdout);

  recurse = force_silent = diagnose_surprises = false;
#ifdef MY_ABC_HERE
  acl_index      = NULL;
  force_keep_acl = false;
#endif //MY_ABC_HERE

  while ((c = getopt_long (argc, argv,
#ifdef MY_ABC_HERE
                           ("Rcfvrk::w::x::X::s::t::u::g::o::a::,::+::=::"
                            "0::1::2::3::4::5::6::7::"),
#else
                           ("Rcfvr::w::x::X::s::t::u::g::o::a::,::+::=::"
                            "0::1::2::3::4::5::6::7::"),
#endif // MY_ABC_HERE
                           long_options, NULL))
         != -1)
    {
      switch (c)
        {
        case 'r':
        case 'w':
        case 'x':
        case 'X':
        case 's':
        case 't':
        case 'u':
        case 'g':
        case 'o':
        case 'a':
        case ',':
        case '+':
        case '=':
        case '0': case '1': case '2': case '3':
        case '4': case '5': case '6': case '7':
          /* Support nonportable uses like "chmod -w", but diagnose
             surprises due to umask confusion.  Even though "--", "--r",
             etc., are valid modes, there is no "case '-'" here since
             getopt_long reserves leading "--" for long options.  */
          {
            /* Allocate a mode string (e.g., "-rwx") by concatenating
               the argument containing this option.  If a previous mode
               string was given, concatenate the previous string, a
               comma, and the new string (e.g., "-s,-rwx").  */

            char const *arg = argv[optind - 1];
            size_t arg_len = strlen (arg);
            size_t mode_comma_len = mode_len + !!mode_len;
            size_t new_mode_len = mode_comma_len + arg_len;
            if (mode_alloc <= new_mode_len)
              {
                mode_alloc = new_mode_len + 1;
                mode = X2REALLOC (mode, &mode_alloc);
              }
            mode[mode_len] = ',';
            memcpy (mode + mode_comma_len, arg, arg_len + 1);
            mode_len = new_mode_len;

            diagnose_surprises = true;
          }
          break;
        case NO_PRESERVE_ROOT:
          preserve_root = false;
          break;
        case PRESERVE_ROOT:
          preserve_root = true;
          break;
        case REFERENCE_FILE_OPTION:
          reference_file = optarg;
          break;
        case 'R':
          recurse = true;
          break;
        case 'c':
          verbosity = V_changes_only;
          break;
#ifdef MY_ABC_HERE
        case 'k':
          force_keep_acl = true;
          break;
#endif // MY_ABC_HERE
        case 'f':
          force_silent = true;
          break;
        case 'v':
          verbosity = V_high;
          break;
        case_GETOPT_HELP_CHAR;
        case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
        default:
          usage (EXIT_FAILURE);
        }
    }

  if (reference_file)
    {
      if (mode)
        {
          error (0, 0, _("cannot combine mode and --reference options"));
          usage (EXIT_FAILURE);
        }
    }
  else
    {
      if (!mode)
        mode = argv[optind++];
#ifdef MY_ABC_HERE
      parse_synoacl_arguments(argc, argv, &mode, &set_acl, &is_acl_opindex);
#endif // MY_ABC_HERE
    }

  if (optind >= argc)
    {
      if (!mode || mode != argv[optind - 1])
        error (0, 0, _("missing operand"));
      else
        error (0, 0, _("missing operand after %s"), quote (argv[argc - 1]));
      usage (EXIT_FAILURE);
    }

  if (reference_file)
    {
      change = mode_create_from_ref (reference_file);
      if (!change)
        die (EXIT_FAILURE, errno, _("failed to get attributes of %s"),
             quoteaf (reference_file));
    }
  else
    {
#ifdef MY_ABC_HERE
      if ( set_acl ) {
        change = mode_compile_acl(mode,
                ( is_acl_opindex && '-' == mode[0] ));
      }
      else
#endif // MY_ABC_HERE
      change = mode_compile (mode);
      if (!change)
        {
          error (0, 0, _("invalid mode: %s"), quote (mode));
          usage (EXIT_FAILURE);
        }
      umask_value = umask (0);
    }

#ifdef MY_ABC_HERE
  if (set_acl && preserve_root)
    {
      error(0, 0, _("acl cannot be set with the option --preserve-root") );
      usage(EXIT_FAILURE);
    }
  if (is_acl_opindex && recurse)
    {
      error(0, 0, _("acl index operation cannot be performed recursively"));
      usage(EXIT_FAILURE);
    }
#endif // MY_ABC_HERE

  if (recurse && preserve_root)
    {
      static struct dev_ino dev_ino_buf;
      root_dev_ino = get_root_dev_ino (&dev_ino_buf);
      if (root_dev_ino == NULL)
        die (EXIT_FAILURE, errno, _("failed to get attributes of %s"),
             quoteaf ("/"));
    }
  else
    {
      root_dev_ino = NULL;
    }

  ok = process_files (argv + optind,
                      FTS_COMFOLLOW | FTS_PHYSICAL | FTS_DEFER_STAT);

  IF_LINT (free (change));

  return ok ? EXIT_SUCCESS : EXIT_FAILURE;
}
